import copy

import tensorflow.keras.layers as tfkl
from tensorflow.python.keras import activations
from tensorflow.python.keras import constraints
from tensorflow.python.keras import initializers
from tensorflow.python.ops import nn

from dynastes import normalizers
from dynastes import regularizers


def _get_regularizers_from_keywords(kwargs):
    _initializers = {}
    _regularizers = {}
    _constraints = {}
    _normalizers = {}

    kwarg_keys = copy.copy(list(kwargs.keys()))

    for kwarg in kwarg_keys:
        if kwarg.endswith('initializer'):
            _initializers[kwarg.split('_initializer')[0]] = initializers.get(kwargs.pop(kwarg, None))
        elif kwarg.endswith('regularizer'):
            if kwarg != 'activity_regularizer':
                _regularizers[kwarg.split('_regularizer')[0]] = regularizers.get(kwargs.pop(kwarg))
        elif kwarg.endswith('constraint'):
            _constraints[kwarg.split('_constraint')[0]] = constraints.get(kwargs.pop(kwarg))
        elif kwarg.endswith('normalizer'):
            _normalizers[kwarg.split('_normalizer')[0]] = normalizers.get(kwargs.pop(kwarg))

    return _initializers, _regularizers, _constraints, _normalizers


class DynastesBaseLayer(tfkl.Layer):
    def __init__(self,
                 activity_regularizer=None,
                 trainable=True,
                 name=None,
                 **kwargs):
        self.weights_dict = {}
        self.initializers, self.regularizers, self.constraints, self.normalizers = _get_regularizers_from_keywords(kwargs)

        super(DynastesBaseLayer, self).__init__(
            trainable=trainable,
            name=name,
            activity_regularizer=regularizers.get(activity_regularizer),
            **kwargs)

    def get_initializer(self, name):
        if name not in self.initializers:
            if name == 'kernel':
                self.initializers['kernel'] = initializers.get('he_uniform')
            elif name == 'bias':
                self.initializers['bias'] = initializers.get('zeros')
            else:
                self.initializers['name'] = initializers.get('glorot_uniform')
        return self.initializers[name]

    def get_regularizer(self, name):
        if name not in self.regularizers:
            self.regularizers[name] = regularizers.get(None)
        return self.regularizers[name]

    def get_constraint(self, name):
        if name not in self.constraints:
            self.constraints[name] = constraints.get(None)
        return self.constraints[name]

    def add_weight(self,
                   name=None,
                   shape=None,
                   trainable=None,
                   partitioner=None,
                   dtype=None,
                   use_resource=None,
                   **kwargs):
        weight = super(DynastesBaseLayer, self).add_weight(name=name,
                                                           shape=shape,
                                                           initializer=self.get_initializer(name),
                                                           regularizer=self.get_regularizer(name),
                                                           trainable=trainable,
                                                           constraint=self.get_constraint(name),
                                                           partitioner=partitioner,
                                                           use_resource=use_resource,
                                                           **kwargs)
        self.weights_dict[name] = weight
        return weight

    def get_weight(self, name, training=None):
        w = self.weights_dict[name]
        if name in self.normalizers:
            w = self.normalizers[name](w, training=training)
        return w

    def get_config(self):
        config = {
            'activity_regularizer': regularizers.serialize(self.activity_regularizer),
        }
        for name, initializer in self.initializers.items():
            config[name + '_initializer'] = initializers.serialize(initializer)
        for name, regularizer in self.regularizers.items():
            config[name + '_regularizer'] = regularizers.serialize(regularizer)
        for name, constraint in self.constraints.items():
            config[name + '_constraint'] = constraints.serialize(constraint)
        for name, normalizer in self.normalizers.items():
            config[name + '_normalizer'] = normalizers.serialize(normalizer)

        base_config = super(DynastesBaseLayer, self).get_config()
        return {**base_config, **config}


class ActivatedKernelBiasBaseLayer(DynastesBaseLayer):
    def __init__(self,
                 activation=None,
                 use_bias=True,
                 activity_regularizer=None,
                 **kwargs):
        super(ActivatedKernelBiasBaseLayer, self).__init__(
            activity_regularizer=regularizers.get(activity_regularizer),
            **kwargs)
        self.activation = activations.get(activation)
        self.use_bias = use_bias

    def build_kernel(self, shape):
        self.add_weight(
            name='kernel',
            shape=shape,
            trainable=True,
            dtype=self.dtype)

    def build_bias(self, output_dim):
        if self.use_bias:
            self.add_weight(
                name='bias',
                shape=(output_dim,),
                trainable=True,
                dtype=self.dtype)

    def call(self, x, training=None):
        if self.use_bias:
            x = nn.bias_add(x, self.get_weight('bias', training=training), data_format='NHWC')
        if self.activation is not None:
            return self.activation(x)
        return x

    def get_config(self):
        config = {
            'activation': activations.serialize(self.activation),
            'use_bias': self.use_bias,
        }
        base_config = super(ActivatedKernelBiasBaseLayer, self).get_config()
        return {**base_config, **config}



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorflow.keras.utils import custom_object_scope
from tensorflow.python.keras import regularizers as tf_regularizers
from tensorflow.python.keras.regularizers import deserialize as _deserialize
from tensorflow.python.keras.regularizers import serialize as _serialize

from . import orthogonal
from .orthogonal import Orthogonal


def serialize(regularizer):
    return _serialize(regularizer)


def deserialize(config, custom_objects={}):
    custom_objects = {**custom_objects, **{'Orthogonal': Orthogonal}}
    return _deserialize(config, custom_objects)


def get(regularizer):
    if type(regularizer) == str:
        if regularizer == 'orthogonal':
            return Orthogonal()
    with custom_object_scope({'Orthogonal': Orthogonal}):
        return tf_regularizers.get(regularizer)


# Cleanup symbols to avoid polluting namespace.
del absolute_import
del division
del print_function



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from dynastes import blocks
from dynastes import layers
from dynastes import normalizers
from dynastes import ops
from dynastes import regularizers
from dynastes import util

__version__ = "0.1.3"
# Cleanup symbols to avoid polluting namespace.
del absolute_import
del division
del print_function



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from .attention_layers import LocalizedAttentionLayer1D
from .attention_layers import LocalizedAttentionLayer2D
from .base_layers import ActivatedKernelBiasBaseLayer
from .time_delay_layers import DepthGroupwiseTimeDelayLayer1D
from .time_delay_layers import DepthGroupwiseTimeDelayLayerFake2D
from .time_delay_layers import TimeDelayLayer1D
from .time_delay_layers import TimeDelayLayerFake2D

# Cleanup symbols to avoid polluting namespace.
del absolute_import
del division
del print_function



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from functools import partial

from tensorflow.python.keras import activations

from dynastes import layers as vqkl
from dynastes.blocks import layer_factory
from dynastes.layers import ActivatedKernelBiasBaseLayer


class LocalizedSelfAttentionBlock1D(ActivatedKernelBiasBaseLayer):

    def __init__(self,
                 attention_dim,
                 output_dim,
                 kernel_size,
                 q_type='Conv1D',
                 k_type='Conv1D',
                 v_type='Conv1D',
                 num_heads=1,
                 multiquery_attention=False,
                 depth_multiplier=1,
                 strides=1,
                 dilation_rate=1,
                 grouped=False,
                 group_size=1,
                 padding='same',
                 activation=None,
                 use_bias=True,
                 **kwargs):
        super(LocalizedSelfAttentionBlock1D, self).__init__(
            activation=activations.get(activation),
            use_bias=use_bias,
            **kwargs)
        self.attention_dim = attention_dim
        self.output_dim = output_dim
        self.kernel_size = kernel_size
        self.num_heads = num_heads
        self.strides = strides
        self.dilation_rate = dilation_rate
        self.padding = padding
        self.activation = activations.get(activation)
        self.use_bias = use_bias
        self.multiquery_attention = multiquery_attention

        conv_partial = partial(layer_factory.get_1d_layer(kernel_size=kernel_size,
                                                          grouped=grouped,
                                                          group_size=group_size,
                                                          depth_multiplier=depth_multiplier,
                                                          padding=padding,
                                                          activation=None,
                                                          use_bias=True,
                                                          kernel_initializer=self.get_initializer('kernel'),
                                                          bias_initializer=self.get_initializer('bias'),
                                                          kernel_regularizer=self.get_regularizer('kernel'),
                                                          bias_regularizer=self.get_regularizer('bias'),
                                                          activity_regularizer=None,
                                                          kernel_constraint=self.get_constraint('kernel'),
                                                          bias_constraint=self.get_constraint('bias')))
        q_filters = attention_dim
        k_filters = attention_dim
        v_filters = output_dim

        if multiquery_attention:
            k_filters /= num_heads
            v_filters /= num_heads

        self.q_layer = conv_partial(type=q_type,
                                    filters=q_filters,
                                    stride=strides,
                                    dilation=dilation_rate, name='Conv-Q')
        self.k_layer = conv_partial(type=k_type,
                                    filters=k_filters,
                                    stride=1,
                                    dilation=1, name='Conv-K')
        self.v_layer = conv_partial(type=v_type,
                                    filters=v_filters,
                                    stride=1,
                                    dilation=1, name='Conv-V')

        attention_padding = padding
        self.attention_layer = vqkl.LocalizedAttentionLayer1D(strides=strides,
                                                              dilation_rate=dilation_rate,
                                                              num_heads=num_heads,
                                                              padding=attention_padding,
                                                              preshaped_q=True)

    def build(self, input_shape):
        self.build_bias(self.output_dim)

    def call(self, inputs, **kwargs):
        x = inputs
        q = self.q_layer(x)
        k = self.k_layer(x)
        v = self.v_layer(x)

        x = self.attention_layer(q=q, k=k, v=v)

        return super(LocalizedSelfAttentionBlock1D, self).call(x, **kwargs)

    def get_config(self):
        config = {
            'attention_dim': self.attention_dim,
            'output_dim': self.output_dim,
            'kernel_size': self.kernel_size,
            'num_heads': self.num_heads,
            'strides': self.strides,
            'dilation_rate': self.dilation_rate,
            'padding': self.padding,
        }
        base_config = super(LocalizedSelfAttentionBlock1D, self).get_config()
        return {**base_config, **config}
